%% deftree.pl: Se pretende que sea un sustituto del actual deftree.pl. 
%% Su labor sera la misma que la del actual con la diferencia de 
%% que tomara su informacion directamente de la salida del parser.
%% (Se elimina la necesidad de la representacion pat para 
%% las reglas y los tipos).

:- module(deftree,	 [defTree/3, call_defTree/2] ).

% Module use.

:- use_module(curryast,  [arity/2, split_rule/3] ).
:- use_module(aux,       [qsort/2, term_at/3] ).
:- use_module(symtab,    [get_rules/2, get_object_property/3] ).

:- use_module(library(lists),      [append/3, length/2, select/3, reverse/2]).
:- use_module(library(aggregates), [findall/3]).

:- include(operators).

%%%% Must CLEANUP 

% Esto es un takeN
comienzos(0, []):-!.
comienzos(N, [[N] | R]):-
	N > 0,
	N1 is N-1,
	comienzos(N1, R).

%%%%%%%%%%%%%%

% Obtains return type of a function.

tipo(A, A) :-
	var(A),
	!.

tipo(A, C) :-
	\+ var(A),
	A = (_A -> B),
	!,
	tipo(B, C).

tipo(A, A) :-
	!.

in(X,[X|_]).
in(X,[_|Xs]) :- 
	in(X,Xs).

distrib_rules(Reglas, Con, Pos, R):- % Parser -> Parser
	in(R,Reglas),
	split_rule(R,Head,_),  
		% Head contiene los parametros de la cabeza en una
	Head1 =.. [dummy|Head],  % lista (sin el nombre del predicado)
	parser2trad(Head1, Head2),
	term_at(Head2, Pos, Term),
	functor(Term, Con_aux, _Arity),
	Con = Con_aux. 

mapAtPos([],_,[]).
mapAtPos([R|Rs], Pos, [T|Ts]) :-
	split_rule(R, Head, _),
	Head1 =.. [dummy | Head],
	parser2trad(Head1, Head2),
	term_at(Head2, Pos, T),
	mapAtPos(Rs, Pos, Ts).

uniform(Reglas,Pos) :-
	mapAtPos(Reglas, Pos, Terms),
	\+ in(var(_), Terms).

casos(_, Casos, Con) :-
	get_object_property(Con, type, Tip),
	tipo(Tip, T_final),
	!,
	findall(Cons-Arity,
	        (
		    get_object_property(Cons, type, Tip2), 
		    tipo(Tip2, T_final),
		    get_object_property(Cons, arity, Arity)
	        ),
		Casos
	       ).

casos(_, [num-1], num) :-
	!.

casos(Reglas, NumsReglas2, X) :-
	number(X),
	lista_nums(Reglas,[],NumsReglas),
	reverse(NumsReglas, NumsReglas2).

% lista_nums/3: Elabora una lista [n1-0, n2-0,...] donde los ni son los
% numeros contenidos en las reglas que se le pasan y 0 representa la aridad de
% las constantes numericas

lista_nums([],Ac,Ac) :- 
	!.
lista_nums([T|Ts],Ac,Nums) :- 
	!,
	lista_nums(T,Ac,N1),
	lista_nums(Ts,N1,Nums).

lista_nums(Num,Ac,[Num-0|Ac]) :-
	number(Num),
	\+member(Num-0,Ac),
	!.

lista_nums(Rule,Ac,Ac1):-
	Rule =.. [:=, Head, _],
	!,
	lista_nums(Head,Ac,Ac1).

lista_nums(Term,Ac,Nums):-
	Term =.. [_|Args],
	!,
	lista_nums(Args,Ac,Nums).

lista_nums(_,Ac,Ac).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% split3(Reglas,Pos,SubCasos).

split3([Regla|Reglas],Pos,Subs) :-
	split_rule(Regla, Head, _),
	Head1 =.. [dummy|Head],
	parser2trad(Head1,Head2),
	term_at(Head2,Pos,Term),
	functor(Term,Functor,_Arity),
	casos([Regla|Reglas],Casos,Functor),
	traduce(Casos,[Regla|Reglas],Pos,Subs).

traduce([],_,_,[]).
traduce([Con-Arity|Casos],Reglas,Pos,[(Con,ReglasCon,NPosPos)|Subs]) :-
	findall(R,distrib_rules(Reglas,Con,Pos,R),ReglasCon),
	posiciones(Arity,NPos),
	prepend(Pos,NPos,NPosPos),
	traduce(Casos,Reglas,Pos,Subs).

posiciones(0,[]):-!.
posiciones(N,[N|Ns]):-
	N1 is N-1,
	posiciones(N1,Ns).

prepend(_,[],[]).
prepend(Pos,[Q|Qs],[PQ|PQs]):-
	append(Pos,[Q],PQ),
	prepend(Pos,Qs,PQs).

%% defTree(Reglas,Posiciones,Arbol)

defTree(Reglas,Poss,branch(Piv,Subs)-Ns):-
	pivote(Reglas,Poss,Piv),
	Reglas=[_|_],
	!,
	Reglas=Ns,  
	split3(Reglas,Piv,Casos), 
	trans(Casos,Piv,Poss,SubAux),
	mapDefTree(SubAux,Subs).

defTree([],_,fallo-[]).

defTree([Regla],Poss,regla-[N]):-
	\+(pivote([Regla],Poss,_)),
	Regla = N.

defTree(Reglas,Poss,or(DefTrees)-Ns):-
	\+ pivote(Reglas,Poss,_),
	Reglas=[_,_|_],
	Reglas = Ns,
	map_regla_defTree(Reglas,Poss,DefTrees).

map_regla_defTree([],_,[]).
map_regla_defTree([Regla|Reglas],Poss,[DefTree|DefTrees]) :-
	defTree([Regla], Poss, DefTree),
	map_regla_defTree(Reglas, Poss, DefTrees).
   
pivote(Reglas,Poss,Piv):-
	in(Piv, Poss),
	uniform(Reglas, Piv).

mapDefTree([],[]).
mapDefTree([(Con,Reglas,Poss)|Ss],[(Con,T)|Ts]) :-
	defTree(Reglas, Poss, T),
	mapDefTree(Ss, Ts).

trans([],_,_,[]).
trans([(Con,Reglas,NPoss)|Casos],Piv,Poss,[(Con,Reglas,NNPoss)|NCasos]) :-
	append(Antes, [Piv|Despues], Poss),
	append(NPoss, Despues, Aux),
	append(Antes, Aux, NNPoss),
	trans(Casos, Piv, Poss, NCasos).

%%%%%%%%%%%%%%%%%%%%% Arboles Paralelos %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

call_defTree(Pred, Arbol):-
	conjuntos_reglas(Pred,Conjuntos_reglas),
	length(Conjuntos_reglas, L),
	L > 1,
	!,
	get_object_property(Pred, type, Type),
	arity(Type,Arity),
	comienzos(Arity, Comienzos),
	findall(A,(
		      member(Conjunto,Conjuntos_reglas),
		      Conjunto = Reglas,
		      defTree(Reglas,Comienzos,A)
		  ),
		  Arboles),
	get_rules(Pred, Ns),
	qsort(Arboles,ArbolesOrd),
	Arbol=or(ArbolesOrd)-Ns.

call_defTree(Pred,Arbol):-
% FIXME
%	lookup(Pred,defType,Type),
%	arity(Type,Arity),
	get_object_property(Pred, arity, Arity),
	comienzos(Arity, Comienzos),
	get_rules(Pred, Reglas),
	defTree(Reglas,Comienzos,Arbol).



%%%%%%%%%%
% This code was in aux before.
% parser2trad(TermPred,TermTrad):- Cambia los subterminos id(A)@Args (Formato
% del parser a A(Args) (formato del traductor). Tambien caambia de id(I) a I

parser2trad(var(X), var(X)) :- 
	!.
parser2trad(id([]),[]) :- 
	!.
parser2trad(id(':')@[A,As],[B|Bs]) :- 
	!,
	parser2trad(A,B),
	parser2trad(As,Bs).

parser2trad(id(P)@Ts,T):-!,
	parser2trad_list(Ts,Ts2),
	T=..[P|Ts2].

parser2trad(id(P),P):-!.
parser2trad(TermP,TermT):-
	TermP=..[Cons|Args],
	Args=[_|_],!,
	parser2trad_list(Args,ArgsT),
	TermT=..[Cons|ArgsT].
parser2trad(T,T).

parser2trad_list([],[]).
parser2trad_list([T|Ts],[T2|Ts2]):-
	parser2trad(T,T2),
	parser2trad_list(Ts,Ts2).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% select(X,[X|Resto],Resto).
% select(X,[Y|Resto],[Y|Z]):-select(X,Resto,Z).

varX_X(var(X),Y,Dicc):-member(X/Y,Dicc),!.
varX_X(Term,TermX,Dicc):-
	Term=..[Id|Args],!,
	varX_X_list(Args,ArgsX,Dicc),
	TermX=..[Id|ArgsX].
varX_X(X,X,_).

varX_X_list([],[],_).
varX_X_list([T|Ts],[Tx|Txs],Dicc):-
	varX_X(T,Tx,Dicc),
	varX_X_list(Ts,Txs,Dicc).

cociente_reglas(Pred,Cociente):-
	% Encuentra todas las Id de regla para pred
        get_rules(Pred, Rules),
	% Las mete a todas las reglas en un lista.
        findall([Id],member(Id,Rules),Cociente0),
	cociente_reglas_aux(Pred,Cociente0,Cociente).

cociente_reglas_aux(_Pred, [], []) :- !.
cociente_reglas_aux(_Pred, [Conj], [Conj]) :- !.
cociente_reglas_aux(Pred, Cociente_ini, Cociente) :-
	% Coge con backtraking una lista de regla
	select(CR1, Cociente_ini, Cociente_ini1),
	select(CR2, Cociente_ini1, Cociente_ini2),
	CR1 = [R1|_],
	CR2 = [R2|_],
	split_rule(R1,H1,_),
	split_rule(R2,H2,_),
	varX_X_list(H1,H1v,Dicc),varX_X_list(H2,H2v,Dicc),
	H1v=H2v,!,
	append(CR1, CR2, Conj1_2),
	append([Conj1_2], Cociente_ini2, Cociente1),
	cociente_reglas_aux(Pred, Cociente1, Cociente).

cociente_reglas_aux(_,Cociente,Cociente).


%%%%%%% Arboles paralelos %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% uno_de_cada(Conjuntos,Conjunto) toma una lista de listas y devuelve con 
% backtracking conjuntos con un elemento de cada uno de los Conjuntos

uno_de_cada([],[]).
uno_de_cada([Lista|Listas],[Elemento|Elementos]):-
	member(Elemento,Lista),uno_de_cada(Listas,Elementos).

% conjuntos_reglas(Predicado,Conjuntos) toma el nombre de un predicado y devuel-
% ve todos los conjuntos de reglas posibles con los que se puede formar un
% arbol determinista. Las reglas se indican por su numero.

conjuntos_reglas(Pred, Conjs):-
	cociente_reglas(Pred,Cociente),
	findall(C,uno_de_cada(Cociente,C),ConjsAux),
	quitar_repetidos(ConjsAux,[],_,ConjsRev),  
	reverse(ConjsRev, Conjs).


quitar_repetidos_lista([],Ac,Ac,[]).
quitar_repetidos_lista([R|Rs],AcIn,AcOut,[R|Rs2]):-
	\+member(R,AcIn),!,
	quitar_repetidos_lista(Rs,[R|AcIn],AcOut,Rs2).
quitar_repetidos_lista([_R|Rs],AcIn,AcOut,Rs2):-
	quitar_repetidos_lista(Rs,AcIn,AcOut,Rs2).

quitar_repetidos([],Ac,Ac,[]).
quitar_repetidos([Lista|Listas],AcIn,AcOut,[Lista2|Listas2]):-
	quitar_repetidos_lista(Lista,AcIn,AcIn2,Lista2),
	quitar_repetidos(Listas,AcIn2,AcOut,Listas2).


%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

