:-module(scanner, [init_scanner/0,skip_blanks/0,makedecimal/1,
        makenumber/1,is_digit/1,remove_char/1,
	token_have/1,token_look/1,token_skip/0,token_mustbe/2,
	token_mustbe/3,resynch/1,get_char/1],[]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-use_module(library(streams)).
:-use_module(library(dynamic)).
:-use_module(library(aggregates)).
:-use_module(library(write)).
%:-use_module(library(iso_byte_char)).
%get0(C) :- get_code(B), char_code(C,B).
:-use_module(my_iso_byte_char).
:-use_module(sicstus).
:-use_module(symtab).

:-dynamic(line_count/1).
:-dynamic(lacs/1).
:-dynamic(symtab/3).
:-dynamic(smode/1).
:-dynamic(token/1).
:-dynamic(opp/3).
:-multifile(line_count/1).
:-multifile(lacs/1).
:-multifile(symtab/3).
:-multifile(smode/1).
:-multifile(token/1).
:-multifile(opp/3).

% egallego@babel: 04 Oct 2004, 16:17
% Este mdulo es el equivalente a un analizador lxico de curry.
% Es usado por el parser.

otherwise:-!.

length([],0):-!.
length([_|Ts],N1):-
	length(Ts,N),
	N1 is N+1.

append([],L,L):-!.
append([H|T],L,[H|L2]):-
	append(T,L,L2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Predicados de acceso a los ultimos caracteres leidos:

init_scanner:-
	retractall(line_count(_)),
	retractall(lacs(_)),
	assert(lacs([])),
	assert(line_count(1)),
	retractall(smode(_)),
	assert(smode(batch)), %%%%%%%%%%% PRUEBA
	retractall(token(_)),
	retractall(symtab(_,_,_)).

incline:-
	retract(line_count(N)),
	N1 is N+1,
	assert(line_count(N1)).

get_char(CH):-
	lacs([CH|_]),!.
get_char(CH):-
	get0(CH),!,
	retract(lacs(List)),
	length(List,N),
	(N<5->List2=[CH|List];
	    append(List3,[_],List),!,List2=[CH|List3]),
	    assert(lacs(List2)).
get_char(eof).

remove_char(Oldest):-
	retract(lacs(List)),
	(List=[] -> assert(lacs([])) ;
	    append(ListIni,[Oldest],List),
	    assert(lacs(ListIni))).

maketoken(Token):-
	get_char(CH),
	transit(CH,Token).

skip_blanks:-
	get_char(CH),
	is_space(CH),!,
	remove_char(CH),char_code(CH,N),
	(N==10->incline;true),
	skip_blanks.
skip_blanks.
  

%%%%%%%%%%% transit/2: Funcion de transicion

transit(CH1,Token):- % -- Comentario
	char_code(CH1,0'-),
	remove_char(_),
	get_char(CH2),char_code(CH2,N2),
	N2==0'-,!,
	remove_char(_),
	skip_comment,
	maketoken(Token).
transit(CH1,Token):- % - symbol
	char_code(CH1,0'-),
	get_char(CH2),
	is_symbol(CH2),!,
	remove_char(_),
	makesym(Sym),
	translate_sym([0'-,CH2|Sym],Token).
transit(CH1,Token):- % - digit
	char_code(CH1,0'-),
	get_char(CH2),
	char_code(CH2,Code2),
	is_digit(CH2),!,
	remove_char(_),
	makenumber(Number),
	char_codes(Number,NumberCodes),
	name(N,[Code2|NumberCodes]),
	Token=num(N).
transit(CH1,op(-)):-char_code(CH1,0'-).

transit(CH1,end_rule):- %%%%%%%%%%%%%%%%%%%%%% PRUEBA
	char_code(CH1,10),
	smode(interactive),!,
	incline,
	remove_char(_).

transit(CH1b,end_rule):- % \n
	char_code(CH1b,10),
	incline,
	remove_char(_), % quitar \n
	get_char(CH1),char_code(CH1,N1),
	(\+is_space(CH1);N1==10),!.
transit(CH1,Token):- % \n
	char_code(CH1,10),!,
	skip_blanks,
	maketoken(Token).
transit(CH,Token):-
	is_space(CH),!,
	get0(CH2),
	remove_char(_),
	transit(CH2,Token).

transit(CH,num(N)):- %digit ...
	is_digit(CH),!,
	remove_char(_),
	makenumber(Number),
	char_codes(Number,NumberCodes),
	char_code(CH,Code),
	name(N,[Code|NumberCodes]).
transit(CH1,Token):-
	char_code(CH1,0''), 
	remove_char(_),
	get_char(CH2),char_code(CH2,N2),
	N2=0'\\,!,
	remove_char(_),
	get_char(CH3),
	remove_char(_),
	get_char(CH4),
	char_code(CH4,N4),
	remove_char(_),
	(N4=0'',trans_char([CH2,CH3],CH)->Token=char(CH);Token=char(error)). 
transit(CH1,Token):-
	char_code(CH1,0''),!,
	get_char(CH2),
	remove_char(_),
	get_char(CH3),char_code(CH3,N3),
	(N3==0'' ->Token=char(CH2);Token=char(error)),
	remove_char(_).    

transit(CH1,str(Token)):- % Cadenas
	char_code(CH1,0'"),
	remove_char(_),
	makestring(Token).

transit(CH,Token3):- % varid o reservedid
	is_lower(CH),!,
	remove_char(_),
	makeid(T2),
	Token2=[CH|T2],
	atom_chars(Token,Token2),
	(is_reserved(Token)->Token3=reserved(Token);(opp(_,_,Token)->Token3=op(Token);Token3=id(Token))).

transit(CH,conid(Token)):-
	is_upper(CH),!,
	remove_char(_),
	makeid(T2),
	Token2=[CH|T2],
	atom_chars(Token,Token2).
						%  name(Token,Token2).

transit(CH1,Token):-
	char_code(CH1,0':),!,
	remove_char(_),
	makesym(CHs),
	List=[CH1|CHs],
	atom_chars(Token2,List),
						%  name(Token2,List),
  (CHs=[]->Token=op(:);Token=consym(Token2)). 

transit(CH,Token):- % varsym -> symbol{symbol|:}
	is_symbol(CH),!,
	remove_char(_),
	makesym(CHs),
	List=[CH|CHs],char_code(CH,N),
	(N==0'\\,smode(interactive),CHs==[],get_char(X),char_code(X,Nx),Nx==10
    ->
	remove_char(_),prompt(_,'>'),maketoken(Token);
	translate_sym(List,Token)).

transit(CH1,X):-
	char_code(CH1,0'(),!,
	remove_char(_), % Quitar el (
	get_char(CH1b),char_code(CH1b,N1),
	(N1==0') -> (X=id(unit),remove_char(CH1b));X='(').  

transit(CH1,id([])):-
	char_code(CH1,0'[),
	remove_char(_), % Quitar el [
	get_char(CH2),char_code(CH2,N2),
	N2==0'],!,
	remove_char(CH2).

transit(CH,Token):-
	is_special(CH),!,char_code(CH,N),
	(N==0'[ ->  true;remove_char(_)), % El [ ya lo has quitado arriba
	char_code(CH,N1),translate_special(N1,Token).

transit(CH1,eof):-
						%  char_code(CH1,Nf),
						%  Nf==(-1),!,
	CH1=eof,!,
	remove_char(_).

%%%%%%%%%%%%%%%%%%%%%% Predicados auxiliares

trans_char([0'\\,0'n],10).
trans_char([0'\\,0'r],13).
trans_char([0'\\,0't],9).
trans_char([0'\\,0'\\],92).
trans_char([0'\\,0'f],12).
trans_char([0'\\,0''],39).

skip_comment:-
	get0(CH),char_code(CH,N),
	(N==10 -> incline ;
	    (N<0 -> fail; skip_comment)).

makestring([]):-
	get_char(CH),
	char_code(CH,N),
	N==0'",!,
	remove_char(_).
makestring([eof]):-
	get_char(CH),
	char_code(CH,N),
	N==(-1),!,
	remove_char(_).
makestring([char(CH)|CHs]):-
	get_char(CH),char_code(CH,N),
	(N==10->incline;true),
	remove_char(CH),
	makestring(CHs).


makeid([]):-
	get_char(CH),
	\+in_id(CH),!.
makeid([eof]):-
	get_char(CH),char_code(CH,N),
	N<0,!,
	remove_char(_).
makeid([CH|CHs]):-
	get_char(CH),
	in_id(CH),!,
	remove_char(_),
	makeid(CHs).

makeid([eof]):-
	get_char(CH),char_code(CH,N),
	N<0,!,
	remove_char(_).
makeid([CH|CHs]):-
	get_char(CH),
	in_id(CH),
	remove_char(_),
	makeid(CHs).

makesym([]):-
	get_char(CH),
	\+in_sym(CH),!.
makesym([CH|CHs]):-
	get_char(CH),
	in_sym(CH),
	remove_char(_),
	makesym(CHs).

translate_sym([0'.,0'.],op(..)):-!.
translate_sym([0':,0':],op(::)):-!.
translate_sym([0':,0'=],op(:=)):-!.
translate_sym([0'=,0'>],op(=>)):-!.
translate_sym([0'>,0'>],op(>>)):-!.
translate_sym([0'>,0'>,0'=],op(>>=)):-!.
translate_sym([0'=],op(=)):-!.
translate_sym([0'=,0'=],op(==)):-!.
translate_sym([0'=,0'=,0'=],op(===)):-!.
translate_sym([0'@],op(@)):-!.
translate_sym([0'\\],op(barra)):-!.
translate_sym(['|'],op(pipe)):-!.
translate_sym([0'<,0'-],op(<-)):-!.
translate_sym([0'-,0'>],op(->)):-!.
translate_sym([0'~],op(tilde)):-!.
translate_sym([0'-],op(-)):-!.
translate_sym([0'+],op(+)):-!.
translate_sym([0'*],op(*)):-!.
translate_sym([0'/],op(/)):-!.
translate_sym([0'>],op(>)):-!.
translate_sym([0'<],op(<)):-!.
translate_sym([0'>],op(>)):-!.
translate_sym([0'>,0'=],op(>=)):-!.
translate_sym([0'=,0'<],op(=<)):-!.
translate_sym([0'&,0'&],op(&&)):-!.
translate_sym([0'/,0'\\],op(/\)):-!.
translate_sym(Tl,op(T)):-!,atom_chars(T,Tl).

translate_special(0'(,'('):-!.
translate_special(0'),')'):-!.
translate_special(0'[,'['):-!.
translate_special(0'],']'):-!.
translate_special(0'{,'{'):-!.
translate_special(0'},'}'):-!.
translate_special(0',,','):-!.
translate_special(0';,';'):-!.
translate_special(0'_,'_'):-!.
translate_special(0'','\''):-!.

in_id(CH):-(is_upper(CH);(is_lower(CH);is_digit(CH))),!.
in_id(0''):-!.
in_id(0'_):-!.
in_id(C):-char_code(C,N),N>127,!.

in_sym(CH):-
	char_code(CH,N),
	member(N,[0':,0'!,0'#,0'$,0'%,0'&,0'*,0'+,0'.,0'/,0'<,0'=,0'>,0'?,0'@,0'\\,0'^,0'|]),!.

is_space(CH):-char_code(CH,N),
	member(N,[9,10,32]),!.
						% is_space(9):-!.
						% is_space(10):-!.
						% is_space(32):-!.

is_digit(CH):-
	char_code(CH,N),
	N<58,
	N>47.

is_upper(CH):-
	char_code(CH,N),
	N>=65,
	N=<90.

is_lower(CH):-
	char_code(CH,N),
	(N>=0'a,N=<0'z ; N=0'_),!.
is_lower(CH):-
	char_code(CH,N),
	N>127,!.

is_symbol(CH):-in_sym(CH).
  
is_special(CH):-
	char_code(CH,N),
	member(N,[0'[,0'],0'(,0'),0',,0';,0'{,0'},0'',0'_]),!.

is_reserved(I):-
	member(I,[and,eval,class,data,default,deriving,do,else,hiding,
	if,import,in,infix,infixl,infixr,instance,let,
	module,newtype,of,or,qualified,rule,then,type,
	where,rigid,flex,curry_eval,choice,
	ord,chr,'String','Char','Bool','Int','Float']),!.

makenumber(Number):-
	makedecimal(Dec1),
	get_char(CH1),
	char_code(CH1,N1),
	(N1=0'.->
	    (remove_char(_),
	    makenumberDpunto(Dec2),
	    append(Dec1,[0'.|Dec2],Number))
	;
	Number=Dec1).

makenumberDpunto(Number):-
	makedecimal(Dec1),
	get_char(CH1),
	(member(CH1,[0'e,0'E]) ->
	    (remove_char(_),
	    makenumberDpuntoDe(Dec2),
	    append(Dec1,[CH1|Dec2],Number))
	;
	Number=Dec1).

makenumberDpuntoDe(Number):-
	get_char(CH1),
	(member(CH1,[0'-,0'+]) ->
	    (remove_char(_),
	    makedecimal(Dec1),
	    Number=[CH1|Dec1])
	;
	(makedecimal(Dec2),Number=Dec2)).

makedecimal([CH1|CHs]):-
	get_char(CH1),
	is_digit(CH1),!,
	remove_char(_),
	makedecimal(CHs).
makedecimal([]).

%%%%%%%%%%%%%%%%%%%%%%%%%%% Predicados de llamada al scanner.

%
%  token_have(Token)
%
%  token_have/1 attempts to get and match the token Token, else fails.
%

token_have(Token):-
	(token(Z) ->
	    Z=Token,
	    retract(token(Z)),! ;
	(otherwise ->
	    maketoken(Y),
	    ( Token = Y -> 
		true ; 
		assert(token(Y)),
		fail ) )).


%
%  token_look(Token)
%
%  token_look/1 Token is the next token in the input (used with next).
%

token_look(Token):-
	(token(Z) -> 
		Z=Token;
	(otherwise ->	
	    maketoken(Y),
	    assert(token(Y)),
	    Token = Y )).


%
% skip(Token)
%
% skip/1  skip the next token in the input.
%

token_skip :-
	token(X) -> retract(token(X)) ; maketoken(_).


%
% token_mustbe(Token,ErrCode)
%
% token_mustbe/2 attempts to get and match the token Token, 
% else gives a syntax_error
%


token_mustbe(Token,ErrCode) :-
	token_have(Token)-> 
		true; 
                (annotate_error(ErrCode),fail). 

token_mustbe(Token,ErrCode,Call) :-
	token_have(Token)->
		true;
		annotate_error(ErrCode),
		call(Call).

%
% syntax_error/0 outputs an error message and the offending token.
%

syntax_error :-
	token(Token) ->
	(Token =.. [_,H|_];Token=H),
	swritef(Error,"%w misplaced.",[H]),!,
	annotate_error(Error).

%
% resynch(List)
%
% Skip tokens until someone belongs to List
%

resynch(List):-
	resynch2([eof|List]).

resynch2(List):-
	token_have(X),
	member(X,List) -> true ; resynch2(List).

%
% skipto(List)
%
%

skipto(List):-
	skipto2([eof|List]).

skipto2(List):-
	token_look(X),
	member(X,List) -> true ; token_skip,skipto2(List). 


% Prueba del scanner:

tokenizar(N):-
	open_input(N,IO),
	tokenizar2,
	close_input(IO).

tokenizar2:-
	init_scanner,
	repeat,
	maketoken(N),
	write(N),write(' '),
	N=eof,!.


