
% Librerias de Sloth

 :-include(operators).
 :-ensure_loaded(dynamics).

 :-use_module(sicstus).
 :-use_module(symtab).
 :-use_module(trad).
 :-use_module(aux).
 :-use_module(scanner).
 :-use_module(typer).
 :-use_module(deftree).
 :-use_module(parser). % OJO!! el lambda-lifting no va

% librerias que se usan para este módulo

:-use_module(library(pretty_print)).
:-use_module(library(lists)).



% definimos un operador infijo asociativo por la derecha ->
 :- op(1200,xfy,->).

% definimos un operador infijo asociativo por la izquierda para aplicacion @
 :- op(1200,yfx,@).

% definimos los predicados dinámicos tipot y reglat
 dynamic(tipot/2).
 dynamic(reglat/3).



%       Extraemos los constructores, data y reglas de la tabla
%       ------------------------------------------------------


% listatipos saca los tipos de los constructores en una lista
listatipos(Rs):- findall((X,Y),lookup(X,type,Y),Rs).

% instipo guarda los tipos de los constructores a partir de listatipos
instipo([]).
instipo([(A,B)|Xs]):- assert(tipot(A,B)), instipo(Xs).


% Tablareglas obtiene la lista de reglas de la tabla de Simbolos (TS)

tablareglas(L):- findall(X,lookup(_,abstree,X),L).

% insregla transforma la lista de reglas obtenidas con tablareglas
% al formato adecuado para ser tipadas, y las guarda.

insregla([]).

insregla([X|Xs]):- X=(A:=B),  A=(A1@R1), verite(B,B1),
                   l2var(regla(A1,R1,B1),Reg,Lvar),
                   annadevars(Lvar,Lvar2,W),
                   sust9(Lvar2,Reg,Reg2),
                   Reg2=regla(RR1,RR2,RR3),
                   assert(reglat(RR1,RR2,RR3)), insregla(Xs).
% por si la regla no respeta el formato.
insregla([X|Xs]):- X\=(A:=B), insregla(Xs). 

% verite comprueba si se trata de un if-then-else (ite).
verite((X->Y),ite(X,A,B)):- Y=(A#B).
verite(X,X):- X\=(A->B).


% el predicado iniciar(X) lo hace todo.
iniciar(X):- listatipos(Rs), instipo(Rs), % obtiene los tipos predefinidos
             tablareglas(L), insregla(L), % obtiene las reglas de la TS
             todaslasreglas(X). % Calcula los tipos de las reglas
             
% falta gtipotabla(X) para guradar las reglas en la TS
% pero por ahora no funciona.





%                     Tipos básicos predefinidos
%                     -------------------------

% tipo Float
  tipo(X,'Float') :- float(X), !.

% tipo Int
  tipo(X,'Int') :- int(X), !.

% tipo List
    tipo([],[_]).
    tipo([X|[]],[Y]) :- tipo(X,Y).
    tipo([X|Xs],[Y]) :- tipo(X,Y), Xs\=[], tipo(Xs,[Y]).

% tipo tupla
    tipo((X,Y),(A,B)) :- tipo(X,A), tipo(Y,B).


 %                 tipos derivados de la tabla de simbolos   
 %                 ---------------------------------------

 tipo(num(N),X):- integer(N),X='Int',! ; X='Float',!.
 tipo(char(_),'Char').
 tipo(str(_),['Char']).



%                             Reglas de inferencia
%                             --------------------

% los tipos sacados de la tabla de simbolos
  tipo(A,T):- tipot(A,T).

% Variables
  tipo(var(X),X).

% condicional
    tipo(ite(A,B,C),T) :- tipo(A,'Bool'), tipo(B,T), tipo(C,T).

% abstracciones
    tipo(lambda(P,E),(T1->T2)):- tipo(P,T1), tipo(E,T2).

    tipo(case([]),X).
    tipo(case([L1|Ls]),(T1->T2)):- tipo(L1,(T1->T2)), tipo(case(Ls),(T1->T2)).

% Aplicaciones
    tipo((F @ X),T2) :- tipo(F,(T1->T2)), tipo(X,T1).

% Let
  tipo(le(X,E1,E2),T) :- tipo(E1,T1), tipo(X,T1), tipo(E2,T).

% Fixed Point
  tipo(fix(X,E),T) :- tipo(X,T), tipo(E,T).




% tipador de reglas

%        convierte las reglas en una sola expresión con lambdas.
%        ------------------------------------------------------


 
 elambdaux(N,S,[],S).
 elambdaux(N,lambda(X,E),[X|Xs],S):- elambdaux(N,E,Xs,S).

 elambda(N,E):-regla(N,L,S), elambdaux(N,E,L,S).
 

% modificamos ecase para que sólo ponga case si hay varias reglas

 ecase(N,case(L)):- setof(X,elambda(N,X),L), \+ hayuna(L).
 ecase(N,X):- setof(X,elambda(N,X),L), hayuna(L),L=[X|[]].

lfix([],[]).
lfix([X|Xs],[Y|R]):- ecase(X,Y), lfix(Xs,R).


% saca la expresión con fix de una componente conexa
% y  sustituye los nombre de reglas por var(X) en los fix,
%  y lo convierte en tuplas.
ffix2(L,L1):- hayuna(L), L=[A], \+esrecur(A), ecase(A,L1),!.
ffix2(L,R):- hayuna(L),L=[A], esrecur(A), ecase(A,B), sust(A,fix(A,B),R),!.
ffix2(L,fix(X,Y)):- \+hayuna(L), lfix(L,L1),
   reemplazo(L,fix(L,L1),R),R=fix(A,B), listatupla(A,X), listatupla(B,Y),!.

% obtiene el tipo de una lista de reglas
tipear([],[]).
tipear([X|Xs],[W,L1]):- ffix2(X,E), tipoerror(E,T), pretty_print(T,[]),
                          gua(X,T,W), tipear(Xs,L1).

% si la expresión es no tipable, se le pone el tipo No Tipable.
tipoerror(E,T):- tipo(E,T),! ; T='No Tipable',!.

% guarda el tipo obtenido por si es necesario para tipear otra regla
gua([A|B],(X,Y),[(A,X),Rs]):- B\=[], assert(tipot(A,X)), gua(B,Y,Rs).
gua([A|[]],T,(A,T)):- assert(tipot(A,T)),!.


% obtiene el tipo de todas las reglas y lo guarda en la lista T
todaslasreglas(T):- ordeneval(L), tipear(L,L1),li(L1,T).


% para guardar la lista de tipos de reglas en la tabla de simbolos      
% OJO, no guarda los tipos en la TS, esta MAL.
gtipotabla([]).
gtipotabla([X|Xs]):- X=(id(A),B), lookup(A,type,B), gtipotabla(Xs).

% obtiene el tipo de una regla (se usa para ejemplos)
tiporegla(X,T):- cconexa(X,L), ffix2(L,E), tipo(E,T), pretty_print(T,[]).




%  predicados para convertir var(x) en var(X) en las reglas         
% ---------------------------------------------------------

  
  % unir une dos listas sin elementos repetidos

   unir([],[],[]):- !.
   unir([],Y,Y):- !.
   unir(X,[],X):- !.
   unir([X|Xs],Y,[X|W]):- noesta(X,Y), unir(Xs,Y,W).
   unir([X|Xs],Y,W):- \+ noesta(X,Y), unir(Xs,Y,W).
   
  % para usar despues de sacar la lista de reglas
   annadevars([],[],var(O)).
   annadevars([var(N)|Rs],[W|Zs],(Y,L)):- copy_term(var(X),Y),
                                      W=(var(N),Y),                                                                 annadevars(Rs,Zs,L). 


  % l2var coge una regla y la transforma en regla valida
  % y saca la lista de variables que tiene.

   l2var(regla(A,B,C),regla(X,Y,Z),L):- l2var(A,X,L1),
                    l2var(B,Y,L2), l2var(C,Z,L3), unir(L1,L2,L4),
                    unir(L4,L3,L).
   l2var(id(A),A,[]).
   l2var(var(A),var(A),[var(A)]).
   l2var((A@B),(X@Y),L):- \+list(B), 
                       l2var(A,X,L1), l2var(B,Y,L2), unir(L1,L2,L).
   l2var((A@B),X,L):- list(B), l2varaux((A@B),C), l2var(C,X,L).


  % para convertir las expresiones A@[ B1 B2 ] en expresiones tipables
  % como ((A@B1)@B2)
   l2varaux((A@[]),A).
   l2varaux((A@[B|Bs]), C):- \+list(B), X=(A@B), l2varaux((X@Bs),C).

   l2var(ite(A,B,C),ite(X,Y,Z),L):- l2var(A,X,L1),
                    l2var(B,Y,L2), l2var(C,Z,L3), unir(L1,L2,L4),
                    unir(L4,L3,L).
   

   l2var(A + B,((@(+,A))@B),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A - B,((@(-,A))@B),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A * B,((@(*,A))@B),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A / B,((@(/,A))@B),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A==B,((@(==,A))@B),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
/*
   l2var(A mod B,@(mod,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A div B,@(div,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A>>B,@(>>,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A>>=B,@(>>=,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A=B,@(=,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
  
   l2var(A<>B,@(<>,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A<B,@(<,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A>B,@(>,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A=<B,@(=<,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A>=B,@(>=,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A&&B,@(&&,@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
   l2var(A'||'B,@('!!',@(X,Y)),L):- l2var(A,X,L1),l2var(B,Y,L2), unir(L1,L2,L).
  
*/ 
   l2var((A,B),(X,Y),L):- l2var(A,X,L1),l2var(B,Y,L2),unir(L1,L2,L).
   l2var(A,X,L):- list(A), ll2var(A,X,L),!.
   l2var(num(X),num(X),[]).
   l2var(char(X),char(X),[]).
   l2var(str(X),str(X),[]).
 
   % para usar l2var con listas
   ll2var([],[],[]):- !.
   ll2var([A|As],[X|Xs],L):- l2var(A,X,L1), ll2var(As,Xs,L2), unir(L1,L2,L).


   
    % sustituye una lista de variables var(x) por sus var(X) correpondientes
    % en la expresión E
    sust9([],E,E).
    sust9([(A,B)|Xs],E,E2):- sust5(A,E,B,E1), sust9(Xs,E1,E2).

  
   %  Sustituye N por S, se usa para sustituir var(x) por var(X) 
   % en las reglas obtenidas por la TS, para que sean tipables.
    sust5(N,N1,S,S):- N==N1.
    sust5(N,regla(A,B,C),S,regla(X,Y,Z)):- sust5(N,A,S,X), sust5(N,B,S,Y),
                                           sust5(N,C,S,Z).
    sust5(N,ite(A,B,C),S,ite(J,K,L)):- sust5(N,A,S,J),sust5(N,B,S,K),
                                       sust5(N,C,S,L).
    sust5(N,(A@B),S,(J@K)):- sust5(N,A,S,J),sust5(N,B,S,K).
    sust5(N,(A,B),S,(J,K)):- sust5(N,A,S,J),sust5(N,B,S,K).
    sust5(N,B,S,B):- atom(B), \+ list(B), N\=B.
    sust5(N,num(X),S,num(X)).
    sust5(N,char(X),S,char(X)).
    sust5(N,str(X),S,str(X)).
    sust5(N,B,S,B):- number(B).
    sust5(N,var(A),S,var(A)):- \+(N==var(A)).
    sust5(N,L,S,L1):- list(L), sustl5(N,L,S,L1).

    % lo mismo que sust5, por si la expresión E es una lista
    sustl5(N,[],S,[]).
    sustl5(N,[X|Xs],S,[A|As]):- sust5(N,X,S,A), sustl5(N,Xs,S,As).
   


%               Calculo de componentes conexas para fix
%               ---------------------------------------


% predicado para ver si un nombre aparece en una expresion

    esta(N,N).
    esta(N,ite(X,Y,Z)):- esta(N,X),! ; esta(N,Y),! ; esta(N,Z).
    esta(N,lambda(X,Y)):- esta(N,X), ! ; esta(N,Y), !.
    esta(N, (X@Y)):- esta(N,X),! ; esta(N,Y),!.
    esta(N,le(X,Y,Z)):- esta(N,X),!; esta(N,Y),! ; esta(N,Z),!.
    esta(N,(X,Y)):- esta(N,X),! ; esta(N,Y), !.

% predicado para sustituir el nombre de la regla por una var(X)
% y usarlo en el tipado de fix (punto fijo)

    sust(N,N,var(X)).
    sust(N,lambda(A,B),lambda(J,K)):- sust(N,A,J), sust(N,B,K).
    sust(N,ite(A,B,C),ite(J,K,L)):- sust(N,A,J),sust(N,B,K),sust(N,C,L).
    sust(N,(A@B),(J@K)):- sust(N,A,J),sust(N,B,K).
    sust(N,le(A,B,C),le(J,K,L)):- sust(N,A,J),sust(N,B,K),sust(N,C,L).
    sust(N,(A,B),(J,K)):- sust(N,A,J),sust(N,B,K).
    sust(N,B,B):- atom(B), \+ list(B), N\=B.
    sust(N,B,B):- number(B).
    sust(N,var(A),var(A)).
    sust(N,num(X),num(X)).
    sust(N,id(X),id(X)).
    sust(N,char(X),char(X)).
    sust(N,str(X),str(X)).

    sust(N,L,L1):- list(L), sustl(N,L,L1).
    sust(N,case(X),case(Y)):- sust(N,X,Y).
    sust(N,fix(X,Y),fix(A,B)):- sust(N,X,A),sust(N,Y,B).

   % como sust, pero para listas
    sustl(N,[],[]).
    sustl(N,[X|Xs],[A|As]):- sust(N,X,A), sustl(N,Xs,As).


 
    % reemplaza todas las variables en un fix([],[]) (A debe ser fix)    
    reemplazo([],A,A).
    reemplazo([X|Xs],A,Z):- sust(X,A,B), reemplazo(Xs,B,Z).

    reemplazo2(fix(A,B),C):- reemplazo(A,fix(A,B),C).


   
% predicado para obtener la expresión final tipable

   efinal(N,fix(N,case(L))):- elambda(N,Y), esta(N,Y),
                           setof(X,elambda(N,X),L),!.

   efinal(N,case(L)):- setof(X,elambda(N,X),L).

% para ver si una función se llama a si misma
esrecur(X):- regla(X,_,Y), esta(X,Y).

% para calcular funcions que llaman a otras funciones
% apunta(N,E) <-- N apunta a E.

apunta(N,E):- regla(N,_,Y), esregla(E), N\=E,  esta(E,Y).

% apunta, si el segundo término es una lista.
% para ver si una componente conexa apunta a una función).
apunta2(N,[A|[]]):- apunta(N,A).
apunta2(N,[X|Xs]):- apunta(N,X), Xs\=[],! ;  apunta2(N,Xs).

% apunta, si el primer término es una lista.
% (para ver si una funcion apunta a una componente conexa).
apunta3([X|[]],Y):- apunta(X,Y).
apunta3([X|Xs],Y):- apunta(X,Y), Xs\=[], ! ; apunta3(Xs,Y).

% apunta si los dos términos son listas (para ver si dos componentes
% conexas se apuntan)
apunta1([X|[]],[Y|Ys]):- apunta2(X,[Y|Ys]).
apunta1([X|Xs],[Y|Ys]):- apunta2(X,[Y|Ys]), Xs\=[],! ; apunta1(Xs,[Y|Ys]).



%                    Predicados para Listas
% algunas se podrían sustituir por las que están en la libreria list
% (programadas en un momento en el que no se disponia de la documentación
% de la libreria list)
%  -----------------------------------------------


% predicado para convertir una lista de listas en una lista
li([X|Xs],[X|Ys]):- \+ list(X), li(Xs,Ys).
li([X|Xs],R):- list(X), li(X,Y), li(Xs,Ys), append(Y,Ys,R).
li([],[]).
 
 
  % convierte una lista en una tupla
  listatupla([A|As],(A,Bs)):- listatupla(As,Bs).
  listatupla([X|[]],X).

 % indica si un elemento NO ESTA en una lista.
 noesta(A,[]).
 noesta(A,[X|Xs]):- A\=X, noesta(A,Xs).

 % elimina un elemento de una lista
  quitar(A,[],[]).
  quitar(A,[A|Xs],Ys):- quitar(A,Xs,Ys).
  quitar(A,[X|Xs],[X|Ys]):- A\=X, quitar(A,Xs,Ys).

 % elimina los elementos de una lista de otra lista
  quitarl([],Xs,Xs).
  quitarl([A|As],Xs,Zs):- quitar(A,Xs,Ys), quitarl(As,Ys,Zs).

 % ordena una lista de funciones que se apuntan unas a otras
  ordenar([],Xs,Xs).
  ordenar([A|As],Bs,Zs):- inserta(A,Bs,Cs), ordenar(As,Cs,Zs).

 % inserta una funcion dependiendo de si apunta o no a las demás
  inserta(A,[],[A]).
  inserta(A,[B|Bs],[A|Cs]):- apunta1(B,A), Cs=[B|Bs].
  inserta(A,[B|Bs],[B|Cs]):- \+ apunta1(B,A), inserta(A,Bs,Cs).

% devuelve una lista sin elementos repetidos
  norep([],[]).
  norep([X|[]],[X|[]]).
  norep([X|Xs],[X|Zs]):- Xs\=[], quitar(X,Xs,Ys), norep(Ys,Zs).

 % devuelve la lista de reglas
  lreglas(R):- findall(X,regla(X,_,_),L), norep(L,R).

% indica si la lista sólo tiene un elemento
  hayuna([X|[]]).



%                Predicados para Reglas
%                ----------------------

  esregla(X):- lreglas(L), pertenece(X,L).
 
  pertenece(X,[X|[]]).
  pertenece(A,[X|Xs]):- A=X ; pertenece(A,Xs).

   

 %funcion para calcular el camino con una lista de grafos

 cam(N,Y):- camaux(N,Y,[]).
 
 camaux(X,X,L).
 camaux(X,Y,L):- apunta(X,Z), noesta(Z,L), camaux(Z,Y,[Z|L]).


 % Cálculo de la componente conexa usando cam
 cconexa2(X,Y):- cam(X,Y), cam(Y,X).

 cconexa(X,L):- setof(Y,cconexa2(X,Y),L).

% calculo de la lista de componentes conexas a partir de la lista de reglas

 lconexa(L):- lreglas(R), lconexa2(R,L).
 lconexa2([],[]).
 lconexa2([X|Xs],[Y|Ys]):- cconexa(X,Y), quitarl(Y,[X|Xs],Zs),
                           lconexa2(Zs,Ys).

% devuelve la lista de componentes conexas ordenadas para su evaluacion

  ordeneval(L):- lconexa(R), ordenar(R,[],L).  






%                             Reglas de la TS
%                            ------------------


   regla(A,B,C):- reglat(A,B,C).






