:- module(lambda, [process_lambdas/0]).

% egallego@babel: 25 Nov 2004, 21:50
% This module rewrites all rules containing lambda(Pattern, Exp) to
% use a new function f' Pattern = Exp

:- use_module(library(dynamic), [assert/1, retract/1]) .
:- use_module(library(terms),   [atom_concat/2] ).

:- include(operators).

:- use_module(symtab,   [get_rules/2, update_rules/2, add_rule/2, set_object_property/3] ).
:- use_module(curryast, [split_rule/3] ).

process_lambdas :-
	get_rules(Name, Rules),
	process_rules(Rules, NewRules),
	update_rules(Name, NewRules),
	fail.

process_lambdas.


process_rules([], []) :- 
	!.
process_rules([Rule|RS], [NewRule | NRS]) :-
	Rule = Head:= Body,
	process_body(Body, NewBody),
	NewRule = Head:=NewBody,
	!,
	process_rules(RS, NRS).

process_body(Rule, NewRule) :-
	Rule = id(X)@L,
	process_body(L, NL),
	NewRule = id(X)@NL.

process_body([Rule|RS], [NR | NRS]) :-
	process_body(Rule, NR),
	process_body(RS,  NRS).
	
process_body(Rule, NewRule@[]) :-
 	Rule = lambda(P, E),
 	!,
 	new_lambda(P, E, NewRule).

process_body(Rule, NewRule) :-
	Rule = lambda(P, E)@L,
	!,
	process_body(L, NL),
	new_lambda(P, E, NewFunction),
	NewRule = NewFunction@NL.


process_body(Rule, Rule).

new_lambda(P, E, id(NewName)) :-
	get_new_name(NewName),
	NewRule = (id(NewName)@P) := E,
	add_rule(NewName, NewRule),
	length(P, Arity),
	set_object_property(NewName, arity, Arity).

:- dynamic lambda_names/1.

get_new_name(Name) :-
	retract(lambda_names(N)),
	!,
	N1 is N + 1,
	assert(lambda_names(N1)),
	atom_number(Num, N),
	atom_concat([lambda, Num], Name).
	
get_new_name('lamdba0') :-
	assert(lambda_names(1)),
	!.

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

