:- module(lexer_compat, 
	[init_scanner/1, token_have/1, token_look/1, token_mustbe/2,
	 token_skip/0, makedecimal/1, resynch/1], _).

%
:- use_module(library(dynamic), [asserta/1, retract/1, retractall/1]).

:- use_module(lexer, [scanFile/2]).

:- dynamic tokens/1.

% Compatibility module.

% Now init_scanner takes an argument, the string to 
% lex by now.
% It works lexing the string and storing it as a predicate
% that token_have will return step by step.

init_scanner(File) :-
	retractall(tokens(_)),
	scanFile(File, Tokens),
	asserta(tokens(Tokens)).

%
%  token_have(Token)
%
%  token_have(Token), gets a token from tokens and updates
%  it.

token_have(Token) :-
	retract(tokens([(Token, _Pos)|TR])),
	asserta(tokens(TR)).

%
%  token_look(Token)
%
%  token_look/1 Token is the next token in the input (used with next).
%

token_look(Token) :-
	tokens([(Token, _Pos)|_]).


%
% skip(Token)
%
% skip/1  skip the next token in the input.
%

token_skip :-
	retract(tokens([_|TR])),
	asserta(tokens(TR)).


makedecimal(N) :-
	token_have(num(N)).

%
% token_mustbe(Token,ErrCode)
%
% token_mustbe/2 attempts to get and match the token Token, 
% else gives a syntax_error
%

:- set_prolog_flag(multi_arity_warnings,off).

token_mustbe(Token,ErrCode) :-
	(
	    token_have(Token)  -> 
	    true
	; 
	    throw(ErrCode),
	    fail
	). 

token_mustbe(Token,ErrCode,Call) :-
	(
	    token_have(Token)  ->
	    true
	;
	    throw(ErrCode),
	    call(Call)
	).

:- set_prolog_flag(multi_arity_warnings,on).

%
% syntax_error is deprecated in this new lexer.
%

syntax_error :-
	throw(syntax_error).

%syntax_error :-
%	token(Token) ->
%	(Token =.. [_,H|_];Token=H),
%	swritef(Error,"%w misplaced.",[H]),!,
%	annotate_error(Error).

%
% resynch(List)
%
% Skip tokens until someone belongs to List
%
resynch(_).

%resynch(List):-
%	resynch2([eof|List]).
%
%resynch2(List):-
%	token_have(X),
%	member(X,List) -> true ; resynch2(List).

%
% skipto(List)
%
%

%skipto(List):-
%	skipto2([eof|List]).

%skipto2(List):-
%	token_look(X),
%	member(X,List) -> true ; token_skip,skipto2(List). 

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

