%:-[init].

:- include(operators).
:- use_module(library(terms_vars), [varset/2] ).

% :- block and_bool(-,?,?),  or_bool(-,?,?).
% :- block req(-,?,?), req(?,-,?).
% :- block apply_list_lazy(-,?,?).
% :- block eval_lazy(-,?).
% :- block '+'(-,?), '+'(?,-), '-'(?,-), '-'(-,?).
% :- block '*'(-,?), '*'(?,-), '/'(?,-), '/'(-,?).

:- multifile constructor/1, tohnf/2, tonf/2, eq/2.
:- discontiguous tohnf/2, tonf/2, contructor/1.

% :-dynamic tohnf/2.


is_in_hnf(A) :-
        (
	    var(A)
	;
	    number(A)
	), 
	!.
is_in_hnf(str(_)) :-
	!.
is_in_hnf(char(_)) :-
	!.
is_in_hnf([]) :- 
	!.
is_in_hnf([_|_]) :- 
	!.
is_in_hnf('Valid') :-
	!.
is_in_hnf(Term) :-
	functor(Term, C, _),
	constructor(C),
	!.

constructor('True').
constructor('False').
constructor(num).
constructor(',').


tohnf(A,A) :- 
	is_in_hnf(A),
	!.

tohnf(_#A, B) :-
        \+ var(A),
	!,
        A = B.

tohnf('Valid','Valid').

tohnf([], []).

tohnf([A|B], [A|B]) :- 
	!.

%tohnf((A,B),(Ahnf,Bhnf)):-!,
%  tohnf(A,Ahnf),tohnf(B,Bhnf).

tohnf((A,B),(A,B)) :-
	!.
tohnf(cond(C,E)#R,R) :- 
	!,
	tohnf(C,Rc),
	Rc == 'Valid',
	tohnf(E,R).

tohnf(done#done,done) :-
	!.
tohnf(return(A)#B,B) :-
	!,
	tohnf(A,B).

tohnf(getChar#char(X), char(X)) :-
	!,
	get0(X).

tohnf(putChar(char(C))#X, Y) :- 
	!,
	X = done,
	Y = done,
	put(C).

tohnf((A>>B)#C,C) :- 
	!,
	tohnf(A,_),
	tohnf(B,C).

tohnf(>>=(A,B)#C,C) :- 
	!,
	tohnf(A,Ahnf),
	B = pap(Right,Remain,List)#_,
	tohnf(apply_list(pap(Right,Remain,List)#_,[Ahnf])#C,C).

tohnf(and_bool(A,C)#B, B) :-
	and_bool(A,C,B).

tohnf(or_bool(A,C)#B, B) :-
	or_bool(A,C,B).

tohnf(not(A)#B, B) :-
        tonf(A, X),
        (
	    X = false ->
	    B = true
	;
	    B = false
	).

tohnf(ifthen(A,B)#C, C) :-
	tonf(A, 'True'),
	tonf(B, C).

tohnf(ifthenelse(A,B,C)#D, D) :-
        (
	    tonf(A, 'True') ->
            tohnf(B, D)
        ;   
	    tohnf(C, D)
        ).

tohnf((D<E)#B,B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	(
	    call((F<G)) -> 
	    B='True' 
	; 
	    B='False'
	).

tohnf((D>E)#B,B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	(
	    call((F>G)) -> 
	    B='True' 
	; 
	    B='False'
	).

tohnf((D>=E)#B,B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	(
	    call((F>=G)) -> 
	    B='True' 
	; 
	    B='False'
	).

tohnf((D=<E)#B,B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	(
	    call((F=<G)) -> 
	    B='True' 
	; 
	    B='False'
	).
 

%tohnf(A#B, B) :-
%        A=..[C,D,E],
%        member(C, [<,>,>=,=<]), !,
%        tonf(D, F),
%        tonf(E, G),
%        H=..[C,F,G],
%        (   call(user:H) ->
%            B='True'
%        ;   B='False'
%        ).

tohnf((A<>B)#X,Y) :-
	X = Y,
	tohnf(A,A1),
	tohnf(B,B1),
	\+ '='(A1,B1),
	!,
	Y = 'True'.

tohnf((A<>B)#X,Y) :-
	X = Y,
	tohnf(A,A1),
	tohnf(B,B1),
	eq(A1,B1),
	!,
	Y = 'False'.

%tohnf(A#B, B) :-
%        A=..[C,D,E],
%        member(C, [+,-,*,/,mod,div]), !,
%        (C=div -> C2='//' ; C=C2),
%        tonf(D, F),
%        tonf(E, G),
%        H=..[C2,F,G],
%        B is H.

tohnf((D+E)#B,B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	B is F+G.

tohnf((D-E)#B,B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	B is F-G.

tohnf((D*E)#B,B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	B is F*G.

tohnf((D/E)#B,B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	B is F/G.

tohnf((D mod E)#B,B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	B is (F mod G).

tohnf(div(D,E)#B,B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	B is F//G.

tohnf(apply_list(A,B)#C, C) :-
        apply_list_lazy(A, B, D),
        eval_lazy(D, C).

tohnf(constraint([])#'Valid','Valid') :-
	!.
tohnf(constraint([C|Cs])#'Valid','Valid') :-
	!,
	tohnf(C,'Valid'),
	tohnf(constraint(Cs)#_,'Valid').

tohnf(choice([(constraint(Const)#_->Expr)#_|Cs])#R,R) :-
	!,
	(
	    tohnf(constraint(Const)#A,A) ->
	    tohnf(Expr,R) 
	; 
	    tohnf(choice(Cs)#R,R)
	).

tohnf_list([],[]).
tohnf_list([A|As],[B|Bs]) :-
	tohnf(A,B),
	tohnf_list(As,Bs).

tonf(A, A) :-
        (
	    number(A)
        ;   
	    var(A)
        ), 
	!.

tonf(char(C),char(C)) :- 
	!.
tonf(A#B,C) :-
	var(A),
	!,
	B = A,
	C = A. %Solo se usa en interactivo.

tonf(done,done) :-
	!.

tonf('False', A) :- 
	!,
        A = 'False'.

tonf('True', A) :- 
	!,
        A = 'True'.

tonf('Valid', A) :- 
	!,
        A = 'Valid'.

tonf((A,B),(Anf,Bnf)) :-
	!,
	tonf(A,Anf),
	tonf(B,Bnf).

tonf([], []) :- 
	!.

tonf([A|B], [C|D]) :- 
	!,
        tonf(A, C),
        tonf(B, D).

tonf(and_bool(A,B),C) :-
	and_bool(A,B,X),
	tonf(X,C).

tonf(or_bool(A,B),C) :-
	or_bool(A,B,X),
	tonf(X,C).

tonf(ifthen(A,B), C) :- 
	!,
        tonf(A, 'True'),
        tonf(B, C).

tonf(ifthenelse(A,B,C), D) :- 
	!,
        (
	    tonf(A, 'True') ->
            tonf(B, D)
        ;   
	    tonf(C, D)
        ).

%tonf(A, B) :-
%        A=..[C,D,E],
%        member(C, [<,>,>=,=<]), !,
%        tonf(D, F),
%        tonf(E, G),
%        H=..[C,F,G],
%        (   call(user:H) ->
%            B='True'
%        ;   B='False'
%        ).
tonf(A<>B,Y) :-
	tohnf(A,A1),
	tohnf(B,B1),
	\+ eq(A1,B1),
	!,
	Y='True'.

tonf(A<>B,Y) :-
	tohnf(A,A1),
	tohnf(B,B1),
	eq(A1,B1),
	!,
	Y='False'.

%tonf(A, B) :-
%        A=..[C,D,E],
%        member(C, [+,-,*,/,mod,div]), !,
%        (C=div -> C2='//' ; C=C2),
%        tonf(D, F),
%        tonf(E, G),
%        H=..[C2,F,G],
%        B is H.

tonf(apply_list(A,B), C) :- 
	!,
        apply_list_lazy(A, B, D),
        eval_lazy(D, C).

tonf(eq(A,B),'True') :- 
	!,
	tohnf(A,At),
	tohnf(B,Bt),
	eq(At,Bt).

tonf(A#B, C) :-
        tohnf(A#B, D),
        tonf(D, C).


eval_lazy(pap(A,0,B)#_, C) :- 
	!,
        revonto_lazy(B, [], D),
        E=..[A|D], % Tambien puede haber constr. currif. (sin puntero)
        (
	    constructor(A)->
	    F =.. [tohnf,E,C]
	;
	    F=..[tohnf,E#C,C]
	),
        call(user:F).

eval_lazy(A, A).

revonto_lazy([], A, A).
revonto_lazy([A|B], C, D) :-
        revonto_lazy(B, [A|C], D).

apply_lazy(pap(A,B,C)#X, D, pap(A,E,[D|C])#X) :-
        E is B-1.

apply_list_lazy(A, [], A).
apply_list_lazy(A, [B|C], D) :-
        apply_lazy(A, B, E),
        apply_list_lazy(E, C, D).


%%%% Igualdad

tohnf(eq(A,B)#'Valid','Valid') :-
	tohnf(A,A1),
	tohnf(B,B1),
	eq(A1,B1).

tohnf('=='(A,B)#C,C) :- % Igualdad estricta (==)
	tohnf(A,A1),
	tohnf(B,B1),
	!,
	'=='(A1,B1,C).

% eq(A,B):-(var(A);var(B)),!,A=B.
eq(A,B) :-
	var(A),
	!,
	tonf(B,A).

eq(A,B) :-
	var(B),
	!,
	tonf(A,B).

eq('True','True') :-
	!.
eq('False','False') :-
	!.
eq([],[]) :- 
	!.
eq([H|Ts],[A|Zs]) :- 
	!,
	tohnf(H,H1),
	tohnf(A,A1),
	eq(H1,A1),
	tohnf(Ts,Ts1),
	tohnf(Zs,Zs1),
	eq(Ts1,Zs1).

eq(A,B) :-
  A =.. [Ca | Alist],
  B =.. [Cb | Blist],
  !,
  Ca = Cb,
  eq(Alist, Blist).

eq(N,M) :-
	number(N),
	!, %% number(M),
	N = M.

%eq(A,B):-
%  A=..[Ca|Alist],B=..[Cb|Blist],!,
%  Ca=Cb,
%  eq(Alist,Blist).


'=='(A, B, 'True') :-
	eq(A, B),
	!.

'=='(A,B,'False') :-
	\+ eq(A, B),
	!.

eq_list([],[]).
eq_list([A|As], [B|Bs]) :-
	eq(A, B),
	eq_list(As, Bs).

and_bool(A, C, B) :-
        tonf(A, X),
        (
	    X == 'False' ->
	    B = 'False'
	;
	    tonf(C,B)
	).

or_bool(A, C, B) :-
        tonf(A, X),
        (
	    X == 'True' ->
	    B =  'True'
	;
	    tonf(C, B)
	).

%%%%%%%%%%%%%%%%%%%%%% Ejecucion del obj en batch: %%%%%%%%%%%%%%%%%

filter_variables_batch([],[]) :-
	!.
filter_variables_batch([loc(_)-_|Vs],Vfs) :-
	!,
	filter_variables_batch(Vs,Vfs).

filter_variables_batch([Var-_|Vs],Vfs) :- % Quitar las var. Rxxx
	name(Var,[82|_]), 
	!,
	filter_variables_batch(Vs,Vfs).

filter_variables_batch([Var-Val|Vs],[Var is Val|Vfs]) :-
	filter_variables_batch(Vs,Vfs).

unfree_vars([],_).

unfree_vars([Atom|Vs],Atom) :-
	unfree_vars(Vs,Atom).

mark_free_vars([]) :- 
	!.
mark_free_vars([Item|Items]) :-
	Item =.. [_,L,R],
	(
	    var(R) ->
	    R = L 
	; 
	    true
	),
	mark_free_vars(Items).

mark_block_vars([]) :- 
	!.
mark_block_vars([L-_|Items]) :-
	unfree_vars(L,'<blocking>'),
	mark_block_vars(Items).

exec_batch(Goal, Dicc) :-
	prompt(_,''),
	Goal = tonf(_Exp,R1),
	varset(Dicc,VarsGoal),
	call_residue(Goal,Res), %comentar?
						%  call(Goal),%descomentar?
        mark_block_vars(Res),                   %comentar?
	mark_free_vars(Dicc),                
	varset(R1,VarsResult),
	unfree_vars(VarsResult,'<free>'),
	write('Result: '),write(R1),nl,
	(
	    VarsGoal == [] ->
	    true
	;
	    write('Substitution: '),
	    write(Dicc),
	    nl,
	    nl,
	    write('More?'),
	    get(Answer),
	    nl,
	    \+ (Answer = 0'; )
        ).


%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

