% Parser de Curry

:- module(parser, [parse/1],[]).

% Awaiting the removal of symtab.
:- include(operators).

% New use_modules clauses, with cleanup.

:- use_module(lexer_compat, [init_scanner/1, token_have/1, token_look/1, token_mustbe/2, 
	                     makedecimal/1,  token_skip/0, resynch/1]).

:- use_module(prelude,      [predefineds/0, forbidden/1]).

:- use_module(symtab,       [add_rule/2, update_rules/2, get_rules/2, get_object_property/3, 
	                     set_object_property/3, add_constructor/2, get_constructors/2]).

:- use_module(curryast,     [arity/2] ).

:- use_module(library(dynamic),    [assert/1] ).
:- use_module(library(lists),      [append/3, length/2, list_concat/2] ).
:- use_module(library(aggregates), [findall/3] ).

% Must kill this, has to do with error handling.
:- use_module(library(write),   [write/1] ).

%

otherwise:-!.

% Dummy error function.
derror(Text) :-
	throw(parser_error(Text)).


% egallego@babel: 06 Oct 2004, 15:54
% Aqu empiezan las reglas gramaticales para el parser.
% Casi todas son del tipo '<noterminal>'
% La regla principal es '<block>'

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Axioma: %%%%%%%%%%%%%%%%%%%%%%%%%%

% Segn se puede ver, un block puede ser:
%  block ::=  
%	EOF
%       END_RULE block
%       DATA type_declaration {; type_declaration} END_RULE block
%       EVAL expr
%       FIXITY fixity_arguments END_RULE block
%       function_name '::' type_expr END_RULE block
%       function_name EVAL annotation END_RULE block
%       function_name equat END_RULE block
% 
% Type classes support.
%       CLASS [context] conid varid WHERE class_members

'<block>' :- token_have(';'), !, '<block>'.
'<block>' :- token_have(eof), !.
'<block>' :-
	token_have(reserved(data)),
	!,
	(
	    '<type_declaration>'(_Name,_Type), % Guardar esta declaracion
	    token_mustbe(';','End of type declarations expected here') -> 
	    true
	;
	    resynch([end_rule])
	),
	'<block>'.

% Preliminary type classes support.
%
%'<block>':-
%	token_have(reserved(class)),!,
%	token_have(conid(ClassName)),
%	token_have(id(ClassParams)),
%	token_have(reserved(where)),
%	'<class_members>'.

%'<class_members>'.

'<block>':-
	token_have(reserved(curry_eval)),
	!,
	(
	    '<expr>'(Expr) ->
	    add_rule('@Objective', Expr)
	;
	    (
		derror(no_goal),
		resynch([end_rule])
	    )
	).

'<block>' :-
	token_look(reserved(FixityKeyword)),
	member(FixityKeyword,[infix,infixl,infixr]),
	!,
	token_skip,
	(
	    '<fixity_arguments>'(Args),
	    token_mustbe(';','End of operator declarations expected here') ->
	    record_fixities(FixityKeyword,Args)
	;
	    resynch([end_rule])
	),
	'<block>'.

'<block>':-
	'<function_name>'(FunctionName),
	(
	    token_have(op('::')) ->
	    ( 
		'<type_expr>'(TypeExpr),
		token_mustbe(';','Start a new decalaratation here, please'),
		record_type(FunctionName,TypeExpr), %registrar la decl. de tipo
		write(FunctionName),
		write('::'),
		write(TypeExpr),
		nl
	    ; % Used as OR
		resynch([end_rule])
	    )
	;
	    (
		'<equat>'(FunctionName, Eq),
		record_equat(FunctionName, Eq) ->
		true
	    ;
		resynch([end_rule])
	    )
	),
	'<block>'.


% Reglas de declaracion de tipos:

% type_expr ::=
%      type_expr_unit 
%      type_expr_unit '->' type_expr
%
%            
%
% TypeExpr ::= TypeExprUnit | TypeExprUnit -> TypeExpr
% TypeExprUnit ::= conid(C) TypeExprUnit1 ... TypeExprUnitN
% TypeExprUnit ::= id(I)
% TypeExprUnit ::= ()
% TypeExprUnit ::= (TypeExpr)
% TypeExprUnit ::= (TypeExpr, ...)
% TypeExprUnit ::= [TypeExpr]
% TypeExprUnit ::= Bool | String | Char |Int | Float

'<type_expr>'(TypeExpr):-
	'<type_expr_unit>'(Tu),
	(
	    token_have(op(->)) ->
	    '<type_expr>'(Te),TypeExpr=(Tu -> Te)
	;
	    TypeExpr=Tu
	).

% type_expr_unit representa la declaracin de un tipo.
% 
% type_expr_unit ::=
%      Constructor {TypeExprUnit}
%      id
%      '(' (type_expr) [','type_expr]  ')'
%      '[' type_expr  ']'
%      Bool | String | Char | Int | Float
%

'<type_expr_unit>'(TypeExpr) :- % Regla 1.
	token_have(conid(C)),
	!,
	'<type_exprs_unit>'(TypeExprList),
	TypeExprAux=id(C)@TypeExprList,
	(
	    TypeExprAux = id('Str') -> 
	    TypeExpr = [id('Char')] 
	; 
	    TypeExpr = TypeExprAux
	).

% egallego@babel
% Type variables must be seen as vars in the typer.
'<type_expr_unit>'(var(I)) :- % Regla 2
	token_have(id(I)),
	!.

'<type_expr_unit>'(TypeExpr):-
	token_have('('),
	!,
	(
	    token_have(')') ->
	    TypeExpr=unit
	;
	    (
		'<type_expr_list>'(TypeListAux),
		(
		    length(TypeListAux,1),
		    TypeListAux = [A] ->
		    TypeExpr=A
		;
		    TypeExpr=tuple(TypeListAux)
		),
		token_mustbe(')',') expected')
	    )
	).

'<type_expr_unit>'(TypeExpr) :-
	token_have('['),
	!,
	'<type_expr>'(TypeAuxLeft),
	token_mustbe(']','] expected'),
	TypeExpr = [TypeAuxLeft].


'<type_exprs_unit>'([T1|Ts]) :-       % <type_exprs_unit>::=
	'<type_expr_unit>'(T1),
	!,                            % lambda | <type_expr_unit><type_exprs_unit>
	'<type_exprs_unit>'(Ts).

'<type_exprs_unit>'([]).

'<type_expr_list>'([T1|Ts]) :- % <type_expr_list>::=
	'<type_expr>'(T1),    % <type_expr> | <type_expr>,<type_expr_list>
	(
	    token_have(',') ->
	    '<type_expr_list>'(Ts)
	;
	    Ts = []
	).

'<type_exprs>'([T1|Ts]) :-       % <type_exprs>::=
	'<type_expr>'(T1),
	!,                       % lambda | <type_expr><type_exprs>
	'<type_exprs>'(Ts).

'<type_exprs>'([]).


'<function_name>'(FunctionName):-
	token_have('('),
	!,
	token_mustbe(op(FunctionName),'Symbol expected'),
	token_mustbe(')',') expected').

'<function_name>'(FunctionName):-
	token_mustbe(id(FunctionName),'A Function Name was expected').

%%%%%%%%%%%%%%%%%%%%%%%%%%%% Declaraciones de tipo: %%%%%%%%%%%%

'<type_declarations>' :- 
	token_have(';'),
	!.

'<type_declarations>' :-
	'<type_declaration>'(_Name,_Type), 
	% Name,Type se registran en type_declaration
	(
	    token_have(';') -> 
	    '<type_declarations>'
	;
	    token_mustbe(';','End of type declarations was expected here')
	).

%
% Estas declaraciones se encuentran en la regla anterior de block:
%       DATA type_declaration {; type_declaration} END_RULE block
%
% type_declaration ::=
%       CONID type_var_ids '=' constr_declaration ['|' constr_declaration]
%

'<type_declaration>'(Name,Type):-
	token_mustbe(conid(Cname),'A constructor name was expected'),
	'<type_var_ids>'(Ids),
	Name =.. [Cname|Ids],
	token_mustbe(op(=),'= expected after type variables'),
	'<constr_declaration>'(ConstrDeclaration),
	(
	    token_have(op(pipe)) -> 
	    '<constr_declarations>'(ConstrDecs),
	    Type = [ConstrDeclaration|ConstrDecs]
	;
	    Type = [ConstrDeclaration]
	),
	(
	    fail -> 
	    true
	; %%% Antes el fail no estaba
	    addconstructors(Type,Name)
	;
	    (
		derror(redefined(Cname))
	    )
	).

%
% Variables de tipo.
%
% type_var_ids ::=
%      { ID }
%

'<type_var_ids>'(List):-!,
	(
	    token_have(id(Id)) ->
	    (
		'<type_var_ids>'(Ids),
		List = [Id|Ids]
	    )
	;
	    List = []
	).  


'<constr_declarations>'(ConstDeclarations):-
	'<constr_declaration>'(ConstDec),
	(
	    token_have(op(pipe)) ->
	    (
		'<constr_declarations>'(ConstDecs),
		ConstDeclarations = [ConstDec|ConstDecs]
	    )
	;
	    (
		ConstDeclarations = [ConstDec]
	    )
	).

% Declaracin de los constructores del tipo de datos.
%
% constr_declaration ::= 
%       CONID type_exprs
%      

'<constr_declaration>'(ConstrDeclaration):-
	token_mustbe(conid(Cname),'A constructor name was expected'),
	'<type_exprs>'(TypeExprs),
	ConstrDeclaration = (Cname,TypeExprs).


%%%%%%%%%%%%%%%%%% Declaracion de Operadores:
%
% block ::=
%       FIXITY fixity_arguments END_RULE block
%
% fixity_argument ::=
%     DECIMAL infix_op {[',' infix_op]}
%
% infix_op ::=
%     OP
%

'<fixity_arguments>'([FixityArg|FixityArgs]):-
	'<fixity_argument>'(FixityArg),
	(
	    token_have(';') -> 
	    '<fixity_arguments>'(FixityArgs) 
	; 
	    FixityArgs = []
	).

'<fixity_argument>'(Nat-Ops):-
	%skip_blanks,
	makedecimal(ListNat),
	name(Nat,ListNat),
	'<infix_op_ids>'(Ops).

'<infix_op_ids>'([op(Op)|Ops]):-
	token_mustbe(op(Op),'Malformed infix operator'),
	(
	    token_have(',') -> 
	    '<infix_op_ids>'(Ops) 
	; 
	    Ops = []
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%% Patrones %%%%%%%%%%%%%%%%%%%%%%%%%%%%

% pattern ::= Constrterm1 ... CostrtermN (N>=0)

'<pattern>'(Pattern):-
	(
	    '<constr_term>'(Constrterm) ->
	    (
		'<pattern>'(Constrterms),
		Pattern = [Constrterm|Constrterms]
	    )
	;
	    Pattern = []
	).

% ConstrTerm ::=
%       VariableID   |  
%       _            | 
%       DataConstrID | 
%       ()           | 
%       (DataconstID ConstrTerm...) | 
%       (ConstrTerm : ConstrTerm)   | 
%       [ConstrTerm, ... ]

'<constr_term_list>'([C|Cs]):-
	'<constr_term>'(C),
	(
	    token_have(',') -> 
	    '<constr_term_list>'(Cs)
	;
	    Cs = []
	).

'<constr_tuple_list>'(X):-
	'<constr_term>'(C),
	(
	    token_have(',') -> 
	    '<constr_tuple_list>'(Cs),
	    X = (C,Cs)
	;
	    X = C
	).

'<constr_term>'(id(I))    :- 
	token_have(id(I)),!.

'<constr_term>'(id(C))    :- 
	token_have(conid(C)),!.

% egallego@babel: Desactivacin temporal.
%'<constr_term>'(id(unit)) :- token_have('('), token_have(')'),!.

'<constr_term>'(id([]))   :- 
	token_have('['), 
	token_have(']'),!.

'<constr_term>'(num(N))   :- 
	token_have(num(N)),
	!.

'<constr_term>'(char(C))  :- 
	token_have(char(C)),!.

'<constr_term>'(S)        :- 
	token_have(str(S)),!.

'<constr_term>'(List):-
	token_have('['),!,
	'<constr_term_list>'(List),
	token_mustbe(']','] expected').

'<constr_term>'(Cons):-
	token_have('('),
	token_have(conid(ID)),
	!,
	(
	    token_have(',') ->
	    (
		'<constr_tuple_list>'(List),
		Cons=(id(ID),List)
	    )
	;
	    (
		token_have(op(:)) ->
		(
		    '<constr_term>'(ConstrTerm),
		    Cons=id(':')@[id(ID),ConstrTerm]
		)
	    ;
		(
		    '<pattern>'(P),
		    Cons = id(ID)@P
		)
	    )
	),
	token_mustbe(')',') expected').

'<constr_term>'(Cons):- % A esta regla solo llegan los ConstrTerm que son
	(
	    token_have(id(X)) ->  %  X donde X no es conid(_)
	    Cons1=id(X) 
	;
	    (
		(
		    token_look(X),
		    member(X,['(','['])
		) -> 
		'<constr_term>'(Cons1)
	    )
	),
	(
	    token_have(',') ->
	    ('<constr_tuple_list>'(List),Cons=(Cons1,List))
	;
	    (
		token_have(op(:)) ->
		(
		    '<constr_term>'(ConstrTerm),
		    Cons=id(':')@[Cons1,ConstrTerm]
		)
	    
	    )
	),
	token_mustbe(')',') expected').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Expresiones: %%%%%%%%%%%%%%%%%%%%%

%
% expr ::=
%     '\' pattern '->' expr
%     IF expr THEN expr ELSE expr
%     '{' constraint 
%     
%     Binary es un poco lo, hay que mirar el tfc para comprender. 
%     Si comprendo bien, en principio una expresin bsica tomara
%     la forma 
%     basic_expr ::=
%         apply
%         basic_expr OP basic_expr
%
%     apply ::=
%         CONID {exprs}
%         FUNID {exprs}
%         


'<expr>'(lambda(P,Expr)):-
	token_have(op(backslash)),!,
	'<pattern>'(P),
	token_mustbe(op(->),'-> expected in lambda expresion'),
	'<expr>'(Expr).

'<expr>'(id(ifthenelse)@[If,Then,Else]):-
	token_have(reserved(if)),
	!,
	'<expr>'(If),
	token_mustbe(reserved(then),'then keyword expected'),
	'<expr>'(Then),
	token_mustbe(reserved(else),'else keyword expected'),
	'<expr>'(Else).

'<expr>'(constraint(Constraint)):-
	token_have('{'),!,
	'<constraint>'(Constraint).

'<expr>'(E1):-
	binary([],6,E1aux),
	(
	    var(E1aux) ->
	    true
	;
	    E1 = E1aux
	).
 
binary(Exc, N, Exp) :-
	(
	    N == 0 ->
	    (
		'<func_type>'(Exp) ->
		true 
	    ;
		token_look(X),
		(
		    X =.. [_,T]
		;
		    X = T
		),
		derror(op_misplaced(T))
	    )
	;
	    (
		otherwise ->
	        N1 is N-1,
		(
		    token_look(op(Op)),
		    opp(N,fx,Op),
		    token_skip ->
		    binary(Exc,N1,Ex),
		    E1 = id(Op)@[Ex],
		    binary4(Exc,N,E1,Exp) 
		;
		    binary(Exc,N1,E1),
		    binary4(Exc,N,E1,Exp)
		)
	    )
	).


binary4(Exc,N,E1,E):-
	(
	    token_look(op(Op)) ->
	    (
		opp(N,Ass,Op),
		member(Ass,[xfx,xfy,yfx]),
		\+ member(Op,Exc) ->
		token_skip,
		make(Exc,N,E1,Ass,Op,E) 
	    ;
		(
		    otherwise ->
		    E1 = E 
		)
	    )
	;
	    (
		otherwise ->
		E1 = E 
	    )
	).


make(Exc, N, E1, xfy, Op, E) :-
	E = id(Op)@[E1,E2],
	binary(Exc,N,E2),
	!.
		
make(Exc, N, E1, yfx, Op, E) :-
	N1 is N-1,
	E3 = id(Op)@[E1,E2],
	binary(Exc,N1,E2),
	binary4(Exc,N,E3,E),
	!.
		
make(Exc, N, E1, xfx, Op, E) :-
	N1 is N-1,
	binary(Exc,N1,E2),
	E = id(Op)@[E1,E2],
	!.

make(_,_,_,_,Op,_):-
	derror(op_misplaced(Op)).
	

'<func_type>'(ExprList):-
	'<basic_expr>'(E1),
	'<func_list>'(Es),
	(
	    Es = [] ->
	    (
		(
		    E1 = id(_)
		;
		    E1 = lambda(_,_)
		) ->
		ExprList = E1@[]
	    ;
		ExprList = E1
	    )
	;
	    ExprList = E1@Es
	).

'<func_list>'(Fl):-
	(
	    '<basic_expr>'(E1) ->
	    (
		'<func_list>'(Es),
		Fl = [E1|Es]
	    ) 
	; 
	    Fl = []
	).

% basic_expr reconoce las expresiones bsicas:
% 
% basic_expr ::= 
%        ID
%        TRUE
%        FALSE
%        tuplas
%        listas
%        
%


'<basic_expr>'(id(X)) :-
	token_have(id(X)),
	!.

'<basic_expr>'(id('True')) :-
	token_have(conid('True')),
	!.
'<basic_expr>'(id('False')) :-
	token_have(conid('False')),
	!.
%'<basic_expr>'(X):-
%  token_have(conid(C)),!,
%  '<func_list>'(List),
%  (List=[]->X=id(C) ; X=id(C)@List). 
'<basic_expr>'(id(C)) :-
	token_have(conid(C)), 
	!.
'<basic_expr>'(char(C)) :- 
	token_have(char(C)),
	!.
'<basic_expr>'(S) :- 
	token_have(str(S)),
	!.
'<basic_expr>'(num(N)) :-
	token_have(num(N)),
	!.
'<basic_expr>'(E):-
	token_have('('),
	!,
	'<expr>'(E1),
	(
	    token_have(',') ->
	    (
		'<expr_tuple_list>'(List),
		E = (E1,List)
	    )
	;
	    E=(E1)
	),
	token_mustbe(')',') expected in expression').

'<basic_expr>'(E):-
	token_have('['),
	!,
	'<expr>'(E1),
	(
	    token_have(',') ->
	    (
		'<expr_list>'(List),
		E=[E1|List]
	    )
	;
	    E = [E1]
	),
	token_mustbe(']','] expected in expresion').

'<expr_list>'([E|Es]):-
	'<expr>'(E),
	(
	    token_have(',') ->
	    '<expr_list>'(Es)
	;
	    Es=[]
	).

'<expr_tuple_list>'(X):-
	'<expr>'(E),
	(
	    token_have(',') ->
	    '<expr_tuple_list>'(Es),
	    X = (E,Es)
	;
	    X = E
	).


%'<basic_constraint>'(BConstraint):- % Esta es la definicion que se atenia a
%  binary([=],6,E1),                 % la gramatica, pero en realidad usan la 
%  token_have(op(=)),                % de abajo:
%  binary([=],6,E2),
%  BConstraint=..[=,E1,E2].

'<basic_constraint>'(BConstraint):-
	'<expr>'(BConstraint).
 
'<constraint>'([]) :-
	token_have('}'),
	!.

'<constraint>'([Bc|Bcs]):-
	'<basic_constraint>'(Bc),
	(
	    token_have(',') -> 
	    '<constraint>'(Bcs)
	;
	    Bcs=[],
	    token_mustbe('}','} expected after constraints')
	).

'<cond_expr>'(cond(constraint(Cond),Expr)) :-
	token_have(op(pipe)),
	token_mustbe('{','{ expected in conditional expression.'),
	'<constraint>'(Cond),
	token_mustbe(op(=),'= expected'),
	'<expr>'(E1),
	(
	    token_have(op(pipe)) ->
	    '<cond_expr>'(CE2),
	    Expr=(E1->CE2)
	;
	    Expr=E1
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'<equat>'(FunctionName,Equat):-
	'<pattern>'(Pattern),
	(
	    token_have(op(=)) -> 
	    (
		'<expr>'(Expr),
		Equat=((id(FunctionName)@Pattern):=Expr)
	    )
	;
	    (
		token_look(op(pipe)) -> 
		(
		    '<cond_expr>'(CondExpr),
		    Equat=((id(FunctionName)@Pattern):=CondExpr))
	    ;
		(
		    derror(pipe_expected),fail)
	    )
	).


% Fixme this should go in prelude.
% opp(1,fx,'not').
opp(1,fx,'-').	
opp(2,xfx,mod).
opp(2,yfx,div).
opp(2,yfx,'*' ). 
opp(2,yfx,'/' ).
opp(3,xfy,':' ).
opp(3,yfx,'-' ).
opp(3,yfx,'+' ). 
opp(4,xfx,'=' ).
opp(4,xfx,'==').
opp(4,xfx,'<' ).
opp(4,xfx,'=<').
opp(4,xfx,'>' ).
opp(4,xfx,'>=').
opp(4,xfx,'<>').
opp(4,xfy,'>>').
opp(4,xfy,'>>=').
opp(5,xfy,'&&' ). 
opp(6,xfy,'||' ). 

      
%% Utilities for recording things in the symbol table.

% Add a list of cons to symbol table.

addconstructors([], _).

addconstructors([(Cons, TypeOfCons) | R], TC) :-
	convert_type(TypeOfCons, TC, Type2),
	arity(Type2, N),
	add_constructor(TC, Cons),
	set_object_property(Cons, type, Type2),
	set_object_property(Cons, arity, N),
	set_object_property(Cons, constructs, TC),
	addconstructors(R, TC).

record_type(Name,Type) :-
	convert_type2(Type,Type2),
	(
	    get_object_property(Name, type, _) 	->
	    derror(dup_type(Name))
	;
	    set_object_property(Name, type, Type2)
	).

record_equat(Name, Eq) :-
	check_arity(Name, Eq), 
	add_rule(Name, Eq).
	

% Debo ver donde meto esto :)

check_arity(Name, Def) :-
	Def = (id(Name)@Params):=_,
	length(Params, Arity),
	(
	    get_object_property(Name, arity, OldArity)->
	    Arity == OldArity
	;
	    set_object_property(Name, arity, Arity)
	).

check_arity(Name, _) :-
	throw(type_error(arity, Name)).


% Recording of new operators.
record_fixities(_,[]).
record_fixities(FixityKeyword,[Nat-Ops|Rs]):-
	record_fixities3(FixityKeyword,Nat,Ops),
	record_fixities(FixityKeyword,Rs).

record_fixities3(_,_,[]).
record_fixities3(Fixity,Nat,[op(Op)|Rs]):-
	member(Fixity-FixProlog,[infix-xfx,infixr-xfy,infixl-yfx]),
	assert(opp(Nat,FixProlog,Op)),
	record_fixities3(Fixity,Nat,Rs).

convert_type([],TC,TC).                 % Convierte el tipo de los constructores al formato
convert_type([id(X)|R], TC, (X->Rc)) :- % que guarda el parser de Jose Andres:
	convert_type(R,TC,Rc).             % Ej: id(Cons)@[id(a),id(Lista)@[id(a)]]
convert_type([id(C)@Args|R],TC,(Cconv->Rconv)):- %pasa a:
	convert_list(Args,ArgsC), % Cons: a -> Lista(a) -> Lista(a)
	Cconv =.. [C|ArgsC],        % (El tipo final lo saca del data)
	convert_type(R,TC,Rconv).
convert_type([tuple(T)|R],TC,(tuple(Tc)->Rc)):-
	convert_list(T,Tc),convert_type(R,TC,Rc).
convert_type([[A]|R],TC,(list(Ac)->Rc)):-
	convert_type2(A,Ac),
	convert_type(R,TC,Rc).

convert_list([],[]).
convert_list([(A->B)|R],[(Ac->Bc)|Rc]):-
	convert_type2(A,Ac),convert_type2(B,Bc),
	convert_list(R,Rc).
convert_list([[A]|R],[Ac|Rc]):-
	convert_type2([A],Ac),
	convert_list(R,Rc).
convert_list([tuple(T)|R],[tuple(Tc)|Rc]):-
	convert_list(T,Tc),convert_list(R,Rc).
convert_list([id(X)|R],[X|Rc]):-
	convert_list(R,Rc).
convert_list([id(X)@Args|R],[Argsc|Rc]):-
	convert_list(Args,Args2),
	Argsc =.. [X|Args2],
	convert_list(R,Rc).

% convert_type2/2: Pasa un tipo tal como lo crea el parser al formato de
% Jose Andres. Ej: id(a) -> id(Tipo)@[id(b)] pasa a:
% a -> 'Tipo'(b)

convert_type2((A->B),(Ac->Bc)):-!,
	convert_type2(A,Ac),convert_type2(B,Bc).
convert_type2(id(A),A):-!.
convert_type2([A],list(B)):-!,
	convert_type2(A,B).
convert_type2(tuple(L),tuple(Lc)):-
	convert_list(L,Lc).
convert_type2(id(A)@List,X):-
	convert_list(List,Listc),
	X=..[A|Listc].


% MUSTCLEAN : Not belongs in parser.
% This is a simple algorithm that changes free ids to prolog vars.

% This first obtains all defined names and then calls cambia_id_var1
cambia_id_var:-
	findall(Name, get_rules(Name, _), NameList),
	findall(ListC, get_constructors(_, ListC), ConsList),
	list_concat(ConsList, ConsList2),
	append(NameList, ConsList2, NL),
	append(NL, [[], unit, '.', mod, div], NL2),
	cambia_id_var1(NL2).

cambia_id_var1(Functions):-
	get_rules(Name, Rules),
	cambia_id_vars_exp_list(Rules, Functions, Abstree_vars),
	update_rules(Name, Abstree_vars),
	fail.



cambia_id_var1(_).

% cambia_id_vars_exp(OriginalExp, ListOfKnownNames, ResultExp).

cambia_id_vars_exp(Var,_,Var):-
	var(Var),!.

cambia_id_vars_exp(id(Id), Fs, X) :-
	!,
	(
	    \+ member(Id, Fs) ->
	    X = var(Id)
	;
	    X = id(Id)
	).

cambia_id_vars_exp(id(Id)@Args, Fs, X):-
	!,
	cambia_id_vars_exp_list(Args, Fs, NArgs),
	(
	    \+ member(Id, Fs) ->
	    (
		NArgs = [] ->
		X = var(Id)
	    ;
		X = var(Id)@NArgs
	    )
	;
	    X = id(Id)@NArgs
	).
cambia_id_vars_exp(Exp,Fs,Exp_c):-
	Exp=..[Op|Args],!,
	cambia_id_vars_exp_list(Args,Fs,Args_c),
	Exp_c=..[Op|Args_c].
cambia_id_vars_exp(Exp,_,Exp).

cambia_id_vars_exp_list([],_,[]).
cambia_id_vars_exp_list([E|Es],Fs,[Ec|Esc]):-
	cambia_id_vars_exp(E,Fs,Ec),
	cambia_id_vars_exp_list(Es,Fs,Esc).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main parsing function.

parse(File):-
	init_scanner(File),
	predefineds,
	'<block>',
	cambia_id_var.

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

