:- module(prolog_writer, [write_rule/1, write_rules/1, write_exp/2] ).

:- use_module(library(write),      [write/1] ).

:- use_module(aux,   [give_name/2]) .

:- use_module(curryast, [varlist/2]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Escritura de las reglas obtenidas.

write_rules([]).
write_rules([Bl|Rs]) :-
	Bl = block(B),
	!,
	write(':-block '),
	write(B),
	write('.'),
	nl,
	write_rules(Rs).

write_rules([R|Rs]) :-
	copy_term(R, R2),  
		% Para evitar que reglas distintas tengan la misma var.
	(
	    R2 = prolog(Head, Body)-_ ->
	    true
	;
	    R2 = prolog(Head,Body)
	),
		% debido a T_Acum (que comparte variables para todas
	varlist(Head, Vhs),
	varlist(Body, Vbs), % las reglas de un predicado).
	append(Vhs, Vbs, Vs),
	give_name(Vs,0),
	write_rule(prolog(Head, Body)),
	write_rules(Rs).


% Escritura de expresiones.

write_exp(A is B, L) :-
	!,
	write_exp(A, L),
	write(' is '),
	write_exp(B, L).

write_exp((A ; B), L) :-
	!,
	write('('),
	write_exp_list(A,L),
	write(';'),
	write_exp_list(B,L),
	write(')').

write_exp([H|T],_) :- 
	!,
	exp2pro([H|T], Tt),
	write(Tt).

% Para el traductor perezoso:

write_exp(ifthen(If,Then),_) :-
	!,
	write('ifthen('),
	write_exp(If,nonl),
	write(','),
	nl,
	write('\t'),
	write_exp(Then,nonl),
	write(')').

write_exp(ifthenelse(If, Then, Else), _) :-
	!,
	write('ifthenelse('),
	write_exp(If,nonl),
	write(','),
	nl,
	write('\t'),
	write_exp(Then,nonl),
	write(','),
	nl,
	write('\t'),
	write_exp(Else,nonl),
	write(')').


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_exp(guardasino(G,Y,N), L) :-
	!,
	write('(('),
	write_exp_list(G,L),
	write(')->'),
	write_exp_list(Y,L),
	write(';'),
	write_exp_list(N,L),write(')').

write_exp(guardasi(G,Y),L):-
	!,write('(('),write_exp_list(G,L),write(')->'),
	write_exp_list(Y,L),write(')').

write_exp(\+(E),L):-!,
	write('\+('),
	write_exp(E,L),
	write(')').

write_exp(var(V1),_):-!,
	copy_term(V1,V),
	(
	    V = loc(Vl) ->
	    write(Vl)
	;
	    (
		V='_' -> 
		true
	    ;
		write('V')
	    ),
	    write(V)
	).
% Si la variable es de las que antes estaba libre, (indicado con loc),
% la escribimos tal cual. Si es del programa ponemos una V delante.

write_exp(num(N),_) :-
	!,
	write_exp(N, _).

write_exp([], _) :- 
	!,
	write('[]').

write_exp(id(I), _) :-
	!,
	name(I, IList),
	IList = [L|_Ls],
	(
	    (
		L > 64,
		L < 91
	    ) ->
	    append([39|IList],[39],NIList)
	;
	    NIList = IList
	),
	name(NI, NIList),
	write(NI).

write_exp(E, L) :-
	E =.. [Op, Left, Right],
	(
	    bin_op(Op)
	;
	    bool_op(Op)
	;
	    Op='#'
	),
	!,
	write('('),           % Para que funcione el traductor perezoso.
	write_exp(Left,L),
	write(' '),write(Op),write(' '),
	write_exp(Right,L),
	write(')').           % Para que funcione el traductor perezoso.

write_exp(str(S), _) :- 
	!,
%  put(39),write(S),put(39),
	write(str(S)).

write_exp(char(C),_) :- 
	!,
	write(char(C)).

write_exp(findall(V, Exps, Res), L) :-
	!,
	write('findall('),
	write_exp(V,L),
	write(',('),
	write_exp_list(Exps,nonl),write('),'),
	write_exp(Res,L),write(')').

write_exp(bagof(Vars,Exist,Exps,Res),L):-!,
	write('bagof('),
	write_exp(Vars,nonl),
	write(','),
	write_tuple(Exist),
	write('('),
	write_exp_list(Exps,noln),
	write('),'),
	write_exp(Res,L),
	write(')').

write_exp(E,_) :-
	E =.. [I|Pars],
	Pars = [_|_],
	!,
	name(I, IList),
	IList = [L | _Ls],
	(
	    (
		L >64,
		L <91
	    ) ->
	    append([39|IList], [39], NIList)
	;
	    NIList = IList
	),
	name(NI,NIList),
	( 
	    L < 64 ->
	    write('\''),
	    write(NI),
	    write('\'') 
	;
	    write(NI)
	),  
	write('('),write_exp_list(Pars,nonl),
	write(')').

write_exp(E, _) :-
	E =.. [I|[]],
	!,
	name(I, IList),
	IList = [L|_Ls],
	(
	    (
		L > 64,
		L < 91
	    ) ->
	    append([39|IList],[39],NIList)
	;
	    NIList = IList
	),
	name(NI, NIList),
	write(NI).

%write_exp(E):-
%	write(E).

write_exp_list([], _) :-
	!.
write_exp_list([H|[]],L) :- 
	!,
	write_exp(H, L).

write_exp_list([H|T], L) :-
	!,
	write_exp(H, L),
	write(','),
	(
	    (
		L = nl
	    ) ->
	    (
		nl,
		write('\t'))
	;
	    write('')
	),
	write_exp_list(T, L).

write_exp_list(E, L) :-
	write_exp(E, L).

write_rule(prolog(H, [])) :-
	!,
	write_exp(H, nonl),
	write('.'),
	nl.

write_rule(prolog(H, T)) :-
	write_exp(H, nonl),
	write(':'),
	write('-'),
	nl,write('\t'),
	write_exp_list(T, nl),
	write('.'),
	nl.

% write_tuple/1 escribe los elementos de una tupla separados por ^

write_tuple((A, B)) :-
	!,
	write_tuple(A),
	write_tuple(B).

write_tuple(var(V)) :-
	write_exp(var(V), nonl),
	write('^').

exp2pro([], []) :-
	!.       % Solo se usa para traducir elementos que
exp2pro([H|T], [Ht|Tt]) :-
	!,       % puedan estar en una lista
	exp2pro(H, Ht),
	exp2pro(T, Tt).

exp2pro(N, N) :-
	number(N),
	!.

exp2pro(id(I), I) :-
	!.
exp2pro(num(N), Nt) :-
	!,
	exp2pro(N,Nt).

exp2pro(var(loc(Vl)), Vl) :-
	!.
exp2pro(var(V), T) :-
	!,
	(
	    V = '_' ->
	    T = V
	;
	    name(V, Vlist),
	    L2 = [86|Vlist],
	    name(T,L2)
	).

exp2pro(E, Et) :-
	E =.. [Op|R],
	name(Op, OpList),
	OpList = [L|_Ls],
	(
	    (
		L > 64,
		L < 91
	    ) ->
	    append([39|OpList],[39],NIList)
	;
	    NIList = OpList
	),
	name(Op2, NIList),
	exp2pro(R, Rt),
	Et =.. [Op2|Rt].

bin_op(X):-
	member(X,[+,-,*,/,mod]),
	!.

bool_op(X):-
	member(X,[=,<>,>=,=<,>,<]),
	!.



%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

